; This program is free software. It comes without any warranty, to
; the extent permitted by applicable law. You can redistribute it
; and/or modify it under the terms of the Do What The Fuck You Want
; To Public License, Version 2, as published by Sam Hocevar. See
; http://sam.zoy.org/wtfpl/COPYING for more details.



GoToEditorOrDisasm:
#ifdef ENABLE_DISASM
	ld	a, (mflags + editorFlags)
	bit	retDisasm, a
	jp	nz, Disassembler
#endif
	jr	HexEditor

;====== Main Stuff =============================================================
hexEditorGoToFreeRAM:
	ld	hl, stringFindFreeRAM
	ld	(editLocation), hl

HexEditor:
	ld	a, (mflags + editorFlags)
	and	retDisasmMask^255
	ld	(mflags + editorFlags), a
EditFullRedraw:
	call	SetFullScrnWind
	call	HomeUp
	ld	hl, (editLocation)
	ld	ixh, textRows-1 ; -1 for index-from-zero, -1 for status bar
editLineLoop:
	ld	bc, (editOffset)
	add	hl, bc
	call	DispHL
	or	a
	sbc	hl, bc
	ld	a, 58 ; colon
	call	PutC
	call	PutSpace
	ld	de, (editPage)
	ld	b, 8
_:	call	GetByte
	inc	hl
	call	DispByte
	bit	0, b
	jr	z, $+7
	push	bc
	call	PutSpace
	pop	bc
	djnz	-_
	call	PutSpace
	ld	b, 8
_:	dec	hl
	djnz	-_
;	ld	b, 4
	ld	b, 8
_:	call	GetByte
	inc	hl
	call	PutC2
	djnz	-_
;	call	PutSpace
;	ld	b, 4
;_:	call	GetByte
;	inc	hl
;	call	PutC2
;	djnz	-_
	dec	ixh
	jr	nz, editLineLoop
	;call	NewLineClrEOL2
	; Status
	call	editShowMemMap
	
;------ Main Editor Loop -------------------------------------------------------
editLoop:
	call	editShowIndics
	call	editShowEffLoc
editLoopNoShowEffLoc:
	call	setCurLoc
	call	CursorOn
	call	GetKey
	call	CursorOff
	ld	de, (editorPage)
	ld	hl, editorKeyJumpTable
	syscall(sysProcessKeyJumpTable)
	jr	editLoop


;------ Status Bar: Memory Map -------------------------------------------------
editShowMemMap:
	ld	hl, textRows-1
	call	Locate
	ld	hl, (editPage)
	call	DispHL
	call	PutSpace
	in	a, (pMPgBHigh)
	call	DispByte
	in	a, (pMPgB)
	call	DispByte
	call	PutSpace
	ld	a, 'R'
	call	PutC
	in	a, (pMPgC)
	add	a, '0'
	call	PutC
	call	PutSpace
	ret

;------ Status Bar: Indicators -------------------------------------------------
editShowIndics:
; 123456789 123456789 123456789 12345
; AAaa BBbb Rr  BC  ----  A DDEE:HHLL
	ld	hl, 13*256 + (textRows-1)
	call	Locate
	call	PutSpace
	ld	a, (mflags + instrFlags)
	bit	noBcall, a
	jr	nz, +_
	ld	a, 'B'
	call	PutC
	ld	a, 'C'
	call	PutC
	jr	++_
_:	call	PutSpace
	call	PutSpace
_:	call	PutSpace
	call	PutSpace
	ld	hl, (editOffset)
	ld	a, l
	or	h
	jr	z, +_
	call	DispHL
	jr	++_
_:	call	PutSpace
	call	PutSpace
	call	PutSpace
	call	PutSpace
_:	call	PutSpace
	ret
	
;------ Status Bar: Effective Location -----------------------------------------
editShowEffLoc:
; 0123456789 123456789 123456789 1234
; AAaa BBbb Rr  BC  ----  A DDEE:HHLL
; AAaa BBbb Rr  BC  ----    R PP:HHLL
	ld	hl, 24*256 + (textRows-1)
	call	Locate
	call	PutSpace
	call	PutSpace
	call	compEffLoc
	ld	a, h
	cp	80h
	jr	nc, editShowEffLocNotPaged
	cp	40h
	jr	c, editShowEffLocNotPaged
	ld	a, 'R'
	ld	de, (editPage)
	bit	7, e
	jr	nz, +_
	ld	a, 'F'
_:	call	PutC
	ld	a, ':'
	call	PutC
	;push	hl
	;ld	hl, (editPage)
	;call	DispHL
	call	CompactPageNumber
	call	DispByte
	;pop	hl
	ld	a, ':'
	call	PutC
	jr	+_
editShowEffLocNotPaged:
	call	PutSpace
	call	PutSpace
	call	PutSpace
	call	PutSpace
	call	PutSpace
_:	;call	compEffLoc
	ld	de, (editOffset)
	add	hl, de
	call	DispHL
	ret

editShowMessage:
	push	hl
	ld	hl, 24*256 + (textRows-1)
	call	Locate
	pop	hl
	ld	de, (editorPage)
	call	PutS
#ifdef ENABLE_DISASM
	ld	a, (mflags + editorFlags)
	bit	retDisasm, a
	jp	nz, disasmLoop
#endif
	jp	editLoopNoShowEffLoc

;------ Cursor Location --------------------------------------------------------
setHexCurLoc:
; Moves the cursor to the location of the current byte under the cursor.
; Destroys:
;  - AF
	; Input: A = {0 1 2 3 4 5 6 7}
	; 0 2 4 6  8 10 12 14
	; 0 2 5 7  10 12 
	; Output: A = {6 8 11 13 16 18 21 23}
	ld	a, (editXLoc)
	ld	h, a
	rr	h
	add	a, a
	add	a, h
	add	a, 6
	and	1Fh
	ld	h, a
	ld	a, (editYLoc)
	ld	l, a
	jp	Locate

setCurLoc:
	ld	a, (editXLoc)
	cp	8
	jr	c, +_
	xor	a
	ld	(editXLoc), a
_:	ld	a, (editYLoc)
	cp	textRows-1
	jr	c, +_
	xor	a
	ld	(editYLoc), a
_:	ld	a, (mflags + editorFlags)
	bit	hexTexsMode, a
	jr	z, setHexCurLoc

setCharCurLoc:
; 123456789 123456789 123456789 12345
; 0123456789 123456789 123456789 1234
; xxxx: 1234 5678 1234 5678  12345678
; yyyy: nnmm oopp qqrr sstt  abcdefgh
	ld	a, (editYLoc)
	ld	l, a
	ld	a, (editXLoc)
;	cp	4
;	ccf
;	adc	a, 26
	add	a, 27
	ld	h, a
	jp	Locate


;------ compEffLoc -------------------------------------------------------------
; Computes the address under the cursor.
; Inputs:
;  - None
; Output:
;  - HL
; Destroys:
;  - AF
compEffLoc:
	ld	a, (editYLoc)
	add	a, a
	add	a, a
	add	a, a
	ld	b, a
	ld	a, (editXLoc)
	add	a, b
	ld	hl, (editLocation)
	add	a, l
	ld	l, a
	ret	nc
	inc	h
	ret


;------ Cursor Movement --------------------------------------------------------
shiftForward:
	ld	hl, (editLocation)
	inc	hl
	ld	(editLocation), hl
	jp	goToEditorOrDisasm

shiftBackward:
	ld	hl, (editLocation)
	dec	hl
	ld	(editLocation), hl
	jp	goToEditorOrDisasm

moveForward:
	ld	a, (editYLoc)
	cp	textRows-1-1
	jr	nz, +_
	ld	a, (editXLoc)
	cp	7
	jr	nz, ++_
	ld	hl, (editLocation)
	ld	de, 64
	add	hl, de
	jr	setLoc
_:	add	a, 8
	cp	textRows-1-1
	jr	nc, +_
	ld	(editYLoc), a
	jp	editLoop
_:	ld	hl, (textRows-1-1)*256 + 7
	ld	(editXLoc), hl
	jp	editLoop

moveLeft:
	ld	a, (editXLoc)
	or	a
	jr	z, +_
	dec	a
	ld	(editXLoc), a
	jp	editLoop
_:	ld	a, 7
	ld	(editXLoc), a
;	jr	moveUp
moveUp:
	ld	a, (editYLoc)
	or	a
	jr	z, +_
	dec	a
	ld	(editYLoc), a
	jp	editLoop
_:	ld	hl, (editLocation)
	ld	de, -8
	add	hl, de
setLoc:	ld	(editLocation), hl
	jp	EditFullRedraw

moveRight:
	ld	a, (editXLoc)
	cp	7
	jr	z, +_
	inc	a
	ld	(editXLoc), a
	jp	editLoop
_:	xor	a
	ld	(editXLoc), a
;	jr	moveDown
moveDown:
	ld	a, (editYLoc)
	cp	textRows-1-1
	jr	z, +_
	inc	a
	ld	(editYLoc), a
	jp	editLoop
_:	ld	hl, (editLocation)
	ld	de, 8
	add	hl, de
	jr	setLoc
	
moveBackward:
	ld	a, (editYLoc)
	or	a
	jr	nz, +_
	ld	a, (editXLoc)
	or	a
	jr	nz, ++_
	ld	hl, (editLocation)
	ld	de, -64
	add	hl, de
	jr	setLoc
_:	sub	8
	jr	c, +_
	ld	(editYLoc), a
	jp	editLoop
_:	ld	hl, 0
	ld	(editXLoc), hl
	jp	editLoop

jumpForward:
	ld	a, (editLocation+1)
	inc	a
	ld	(editLocation+1), a
	jp	goToEditorOrDisasm

jumpBackward:
	ld	a, (editLocation+1)
	dec	a
	ld	(editLocation+1), a
	jp	goToEditorOrDisasm

bigJumpForward:
	ld	a, (editLocation+1)
	add	a, 10h
	ld	(editLocation+1), a
	jp	goToEditorOrDisasm

bigJumpBackward:
	ld	a, (editLocation+1)
	sub	10h
	ld	(editLocation+1), a
	jp	goToEditorOrDisasm

toggleMode:
	ld	a, (mflags + editorFlags)
	xor	hexTexsModeMask
	ld	(mflags + editorFlags), a
	bit	hexTexsMode, a
	jp	nz, goToEditorOrDisasm
	call	ClearAllShiftFlags
	jp	goToEditorOrDisasm


;------ Go to Address ----------------------------------------------------------
gotoAddress:
; 123456789 123456789 123456789 12
; AAaa BBbb Rr BC ---- A DDEE:HHLL
	call	stashCurrentAddress
	ld	hl, 0
	call	Locate
	call	GetHexByte
	push	af
	call	GetHexByte
	ld	l, a
	pop	af
	ld	h, a
	ld	de, (editOffset)
	or	a
	sbc	hl, de
	ld	(editLocation), hl
	jp	goToEditorOrDisasm

hexEditPage:
	call	stashCurrentAddress
;	in	a, (pIntId)
;	and	8
;	jr	z, +_
	ld	hl, textRows-1
	call	Locate
	call	GetHexByte
	push	af
	call	GetHexByte
	ld	l, a
	pop	af
	ld	h, a
	ld	(editPage), hl
	jp	goToEditorOrDisasm
;_:	ld	hl, 5*256 + textRows-1
;	call	Locate
;	call	GetHexByte
;	out	(pMPgBHigh), a
;	call	GetHexByte
;	out	(pMPgB), a
;	jp	goToEditorOrDisasm


; 123456789 123456789 123456789 12
; 0123456789 123456789 123456789 1
; AAaa BBbb Rr BC ---- A No more.
; AAaa BBbb Rr BC ---- A Bookmarkd
; AAaa BBbb Rr BC ---- A Find:1234

;------ Editing ----------------------------------------------------------------
editByte:
	ld	hl, 24*256 + (textRows-1)
	call	Locate
	ld	de, (editorPage)
	ld	hl, strWrite
	call	PutS
	call	setCurLoc
	ld	a, (mflags + editorFlags)
	bit	hexTexsMode, a
	jr	nz, +_
	call	GetHexByte
	jr	editByteCheck
_:	ld	a, allowAlphaMask | allowLwrAlphaMask | allowNumbersMask | allowSymbolsMask
	ld	(mflags + inputerFlags), a
	call	CursorOn
	call	GetAscii
	push	bc
	call	CursorOff
	pop	af
editByteCheck:
	push	af
	call	compEffLoc
	pop	af
	ld	de, (editPage)
	call	PutByte
	push	hl
	call	setHexCurLoc
	pop	hl
	ld	de, (editPage)
	call	GetByte
	push	af
	call	DispByte
	call	setCharCurLoc
	pop	af
	call	PutC2
	jp	editLoop

;------ Write-a-Byte & Erase ---------------------------------------------------
editConfirmFlash:
	ld	de, (editorPage)
	ld	hl, strAddress
	call	PutS
	ld	de, (editPage)
	ld	a, d
	call	DispByte
	ld	a, e
	call	DispByte
	ld	a, d
	cp	80h
	jr	nc, +_
	call	compEffLoc
	ld	a, h
	cp	swapLoc >> 8
	jr	c, +_
	cp	swapTop >> 8
	jr	c, ++_
_:	call	NewLineClrEOL
	ld	de, (editorPage)
	ld	hl, strNotFlashZone
	call	PutSClrWind
	call	GetKey
	pop	af	; Stack
	jp	goToEditorOrDisasm
_:	call	DispHL
	ld	hl, (currentRow)
	push	hl
	ld	de, (editorPage)
	ld	hl, strNewData-1
	call	PutSClrWind
	pop	hl
	call	Locate
	jp	NewLine


editFlashWrite:
	call	HomeUp
	ld	de, (editorPage)
	ld	hl, strFlashWriteBanner
	call	PutS
	call	NewLineClrEOL
	call	editConfirmFlash
	ld	de, (editorPage)
	ld	hl, strNewData
	call	PutS
	call	GetHexByte
	push	af
	call	NewLine
	ld	de, (editorPage)
	ld	hl, strWarning
	call	PutS
	call	GetKey
	cp	skPrgm
	jr	z, +_
	pop	af
	jp	goToEditorOrDisasm
_:	ld	de, (editorPage)
	ld	hl, strClobbering
	call	PutS
	call	compEffLoc
	ld	de, (editPage)
	pop	bc
	call	WriteFlashByte
	ld	de, (editorPage)
	ld	hl, strDone
	call	PutS
	call	GetKey
	jp	goToEditorOrDisasm


editFlashErase:
	call	HomeUp
	ld	de, (editorPage)
	ld	hl, strFlashErase
	call	PutS
	call	NewLineClrEOL
	call	editConfirmFlash
	ld	de, (editorPage)
	ld	hl, strWarning
	call	PutS
	call	GetKey
	cp	skPrgm
	jp	nz, goToEditorOrDisasm
	ld	de, (editorPage)
	ld	hl, strClobbering
	call	PutS
	call	compEffLoc
	ld	de, (editPage)
	call	EraseSector
	ld	de, (editorPage)
	ld	hl, strDone
	call	PutS
	call	GetKey
	jp	goToEditorOrDisasm


;------ Flash Block Operations -------------------------------------------------
editFlashBlockOp:
	ld	hl, mFlashBlockOp
	jp	Menu
editFlashBlockWrite:
	call	HomeUp
	ld	de, (editorPage)
	ld	hl, strFlashWrite
	call	PutS
	call	NewLineClrEOL
	scf
	jr	editBlockDo
editBlockCopy:
	call	HomeUp
	ld	de, (editorPage)
	ld	hl, strBlockCopy
	call	PutS
	call	NewLineClrEOL
	or	a
editBlockDo:
	push	af
	ld	de, (editorPage)
	ld	hl, strSourceAddress
	call	PutS
	ld	hl, (currentRow)
		push	hl
		ld	de, (editorPage)
		ld	hl, strNewData-1
		call	PutSClrWind
		pop	hl
	call	Locate
	; Source
	call	GetHexByte
		push	af
		call	GetHexByte
		ld	e, a
		pop	af
	ld	d, a
	push	de
	call	GetHexByte
		push	af
		call	GetHexByte
		ld	l, a
		pop	af
	ld	h, a
	push	hl
	call	NewLine
	; Destination
	ld	de, (editorPage)
	ld	hl, strDestinationAddress
	call	PutS
	call	GetHexByte
		push	af
		call	GetHexByte
		ld	e, a
		pop	af
	ld	d, a
	push	de
	call	GetHexByte
		push	af
		call	GetHexByte
		ld	l, a
		pop	af
	ld	h, a
	push	hl
	call	NewLine
	; Size
	ld	de, (editorPage)
	ld	hl, strBytesToCopy
	call	PutS
	call	GetHexByte
		push	af
		call	GetHexByte
		ld	e, a
		pop	af
	ld	d, a
	push	de
	call	NewLine
	ld	de, (editorPage)
	ld	hl, strWarning
	call	PutS
	call	GetKey
	cp	skPrgm
	jr	z, +_
	pop	af
	pop	af
	pop	af
	pop	af
	pop	af
	pop	af
	jp	goToEditorOrDisasm
_:	ld	de, (editorPage)
	ld	hl, strClobbering
	call	PutS
	pop	ix
	exx
	pop	hl
	pop	de
	exx
	pop	hl
	pop	de
	exx
	pop	af
	ld	a, 0
	adc	a, a
	ld	b, a
	exx
	di
	call	BlockCopy
	ei
	ld	de, (editorPage)
	ld	hl, strDone
	call	PutS
	call	GetKey
	jp	goToEditorOrDisasm


;------ Bookmarks --------------------------------------------------------------

stashCurrentAddress:
; Stores the current page/address mapping into bookmark 0.
; Destorys:
;  - AF, BC, HL
	ld	hl, editBookmark0
storeCurrentAddress:
; Writes the current page/address mapping into the given bookmark struct
; Input:
;  - HL: Pointer to struct to write to
; Destroys:
;  - AF
;  - BC
;  - HL
	ld	bc, (editLocation)
	ld	(hl), c
	inc	hl
	ld	(hl), b
	inc	hl
	ld	bc, (editPage)
	ld	(hl), c
	inc	hl
	ld	(hl), b
	inc	hl
	ld	bc, (editXLoc)
	ld	(hl), c
	inc	hl
	ld	(hl), b
	ret

exchangeCurrentAddressWithBookmark0:
; Swaps the current location with the location in bookmark 0.
; Uses the stack for scrap.
; Destorys:
;  - AF, BC, DE, HL
	ld	hl, editBookmark0+sizeOfBkm-1
	ld	d, (hl)
	dec	hl
	ld	e, (hl)
	dec	hl
	push	de
	ld	d, (hl)
	dec	hl
	ld	e, (hl)
	dec	hl
	push	de
	ld	d, (hl)
	dec	hl
	ld	e, (hl)
	push	de
	ld	hl, 0
	add	hl, sp
	call	loadBookmark
	pop	af
	pop	af
	pop	af
	jp	goToEditorOrDisasm

loadBookmark:
; Input:
;  - HL: Pointer to bookmark struct
	push	hl
	call	stashCurrentAddress
	pop	hl
	ld	e, (hl)
	inc	hl
	ld	d, (hl)
	inc	hl
	ld	(editLocation), de
	ld	a, d
	cp	40h
	jr	c, restoreAddrNoRemap
	cp	080h
	jr	nc, restoreAddrNoRemap
	ld	e, (hl)
	inc	hl
	ld	d, (hl)
	inc	hl
	ld	(editPage), de
	jr	+_
restoreAddrNoRemap:
	inc	hl
	inc	hl
_:	ld	a, (hl)
	inc	hl
	cp	textCols
	jr	nc, +_
	ld	(editXLoc), a
_:	ld	a, (hl)
	inc	hl
	cp	textRows
	ret	nc
	ld	(editYLoc), a
	ret

goBookmark9:
	inc	a
goBookmark8:
	inc	a
goBookmark7:
	inc	a
goBookmark6:
	inc	a
goBookmark5:
	inc	a
goBookmark4:
	inc	a
goBookmark3:
	inc	a
goBookmark2:
	inc	a
goBookmark1:
	ld	hl, editBookmarks
#if sizeOfBkm != 6
	.error	"Code in goBookmarkN must be adjusted to reflect correct size of bookmark struct."
#endif
	add	a, a
	ld	b, a
	add	a, b
	add	a, b
	add	a, l
	ld	l, a
	jr	nc, +_
	inc	h
_:	in	a, (pIntId)
	and	8
	jr	nz, +_
	; Save
	call	storeCurrentAddress
	; Show message
	ld	hl, strBookmarkSaved
	jp	editShowMessage
_:	call	loadBookmark
	jp	goToEditorOrDisasm


;------ Quick Find -------------------------------------------------------------

editFind8:
	ld	hl, editFind8UiBlock
	call	editVal
	ld	a, (mflags + editorFlags)
	and	findQuick16Mask^255
	ld	(mflags + editorFlags), a
	jr	editQuickFindDo
editFind16:
	ld	hl, editFind16UiBlock
	call	editVal
	ld	a, (mflags + editorFlags)
	or	findQuick16Mask
	ld	(mflags + editorFlags), a

editQuickFindDo:
	call	stashCurrentAddress
	call	compEffLoc
	ld	de, (editQuickVal)
	ld	a, (mflags + editorFlags)
	ld	b, a
editQuickFindDoLoop:
	inc	hl
	ld	a, (hl)
	cp	d
	jr	z, +_
	ld	a, l
	or	h
	jr	nz, editQuickFindDoLoop
	call	editShowMemMap
	call	editShowIndics
	ld	hl, strNoMore
	jp	editShowMessage
;	jp	editLoopNoShowEffLoc
_:	bit	findQuick16, b
	jr	z, +_
	inc	hl
	ld	a, (hl)
	dec	hl
	cp	e
	jr	nz, editQuickFindDoLoop
_:	ld	(editLocation), hl
	ld	hl, 0
	ld	(editXLoc), hl
	jp	goToEditorOrDisasm


;------ Options ----------------------------------------------------------------
; Editable options:
;  - Address (hotkey)
;  - Page (hotkey)
;  - Offset (hotkey)
;  - Find location
;  - Wildcard existance
;  - Wildcard value
;  - Find8 val
;  - Find16 val
;New key mappings:
;F3: Flash options? Autoselect tests?
;Mode: Toggle hex/char mode
;Del: Redraw
;XTThn: Set offset such that the given value is displayed as the current address
;On+XTthn: Set offset manually
;Prgm: Set find string location to here
;On+prgm: Set find string location manually
;Vars: Go to find string
;sin: Find 16
;cos: Find 8
;On+^: Set B page
;,: Execute string find
;On+log: Do flash erase on current location
;ln: Block copy
;On+ln: Block flash write
;sto: Write-a-byte
;
;Open keys:
;.
;^-1
;x^2
;Apps
;Math
;Alpha
;2nd
;F1
;F2

;hexEditMenu:
;	ld	hl, mHexEditOptions
;	jp	Menu

editVal:
	push	hl
	ld	hl, textRows-1
	call	Locate
	pop	hl
	ld	de, (editorPage)
	call	PutS
	push	hl
	ld	hl, (currentRow)
	push	hl
	call	NewLineClrEOL
	pop	hl
	call	Locate
	pop	hl
	ld	b, (hl)
	inc	hl
	ldHlIndirect
_:	call	GetHexByte
	ld	(hl), a
	dec	hl
	djnz	-_
	call	editShowMemMap
	call	editShowIndics
	ret

editChangeOffset:
	in	a, (pIntId)
	and	8
	jr	nz, +_
	ld	de, (editLocation)
	ld	hl, (editOffset)
	or	a
	sbc	hl, de
	ld	(editOffset), hl
	jp	goToEditorOrDisasm
_:	ld	hl, editChangeOffsetUiBlock
	call	editVal
	jp	goToEditorOrDisasm

	

;------ String Find ------------------------------------------------------------
editSetWildcard:
	ld	a, (mflags + editorFlags)
	xor	findUseWildcardMask
	ld	(mflags + editorFlags), a
	bit	findUseWildcard, a
	jr	z, +_
	ld	hl, editWildcardUiBlock
	call	editVal
	jp	editLoop
_:	ld	hl, strWildcardOff
	jp	editShowMessage		

editStringSetLoc:
	call	compEffLoc
	ld	(editFindSource), hl
	ld	de, (editPage)
	ld	(editFindPage), de
	ld	hl, strFindLocSet
	jp	editShowMessage	

editStringSetLen:
	ld	a, (mflags + editorFlags)
	bit	findManLen, a
	jr	z, +_
	res	findManLen, a
	ld	(mflags + editorFlags), a
	ld	hl, strFindAutoLen
	jp	editShowMessage
_:	set	findManLen, a
	ld	(mflags + editorFlags), a
	ld	hl, editStrFindLength
	call	editVal
	jp	editLoop
	
doStringFind:
	call	compEffLoc
	ld	de, (editPage)
	exx
	ld	hl, (editFindSource)
	ld	de, (editFindPage)
	ld	a, (mflags + editorFlags)	; alternate
	bit	findManLen, a
	jr	nz, stringFindManLen
	ld	ixl, 0
	; Find 0 term
_:	inc	ixl
	call	IncPtr
	call	GetByte
	or	a
	jr	z, -_
	jr	+_
stringFindManLen:
	; Get manual length
	ld	a, (editFindLength)
	ld	ixl, a
_:	ld	a, (editWildcard)
	ld	ixh, a	
	ld	hl, (editFindSource)
	ld	de, (editFindPage)
	ld	a, (mflags + editorFlags)
	ld	c, a
	exx
	call	StringFind
	jp	nz, goToEditorOrDisasm
	ld	(editPage), de
	exx
	call	compEffLoc
	ld	de, (editLocation)
	or	a
	sbc	hl, de
	push	hl
	exx
	pop	de
	or	a
	sbc	hl, de
	ld	(editLocation), hl
	jp	goToEditorOrDisasm
StringFind:
; Finds a string.
; Inputs:
;  - DEHL: Start location-1
;  - DEHL': String location
;  - C': 1 if using wildcard
;  - IXH: Wildcard
;  - IXL: String length
; Outputs:
;  - 
	exx					; 
	ld	a, (editWildcard)		; alternate
	ld	ixh, a				; alternate
	call	GetByte				; alternate
	exx					; 
	ld	b, a				; 
; Registers:
;  - DEHL: Current search location
;  - DEHL': String search location
;  - B: First char in string
;  - C: Swap port
;  - IXH: Wildcard
;  - IXL: String length
;  - C': Stuff
stringFindLoop:
	in	a, (pIntId)
	and	intOnKeyDown
	ret	z
	; inc ptr
	call	IncPtr				; 
	; check for end
	bit	7, e				; 
	jr	nz, +_				; 
	bit	1, d				; 
	ret	nz				; 
	jr	++_				; 
_:	bit	3, e				; 
	ret	nz				; 
_:	call	GetByte				; 
	cp	b				; 
	jr	nz, stringFindLoop		; 
	push	de				; 
	push	hl				; 
	exx					; 
	push	de				; alternate
	push	hl				; alternate
	push	ix				; alternate
stringFindTestLoop:
	in	a, (pIntId)			; alternate
	and	intOnKeyDown			; alternate
	ret	z				; alternate
	call	GetByte				; alternate
	exx					; 
	ld	b, a				; 
	call	GetByte				; 
	bit	findUseWildcard, c		; 
	jr	z, +_				; 
	cp	ixh				; 
	jr	z, ++_				; 
_:	cp	b				; 
	jr	nz, stringFindTestFailed	; 
_:	call	IncPtr				; 
	exx					; 
	call	IncPtr				; alternate
	dec	ixl				; alternate
	jr	nz, stringFindTestLoop		; alternate
	; Found! Fix registers.
	pop	ix				; alternate
	pop	hl				; alternate
	pop	de				; alternate
	exx					; 
	pop	hl				; 
	pop	de				; 
	ret					; 
stringFindTestFailed:
	exx					; 
	pop	ix				; alternate
	pop	hl				; alternate
	pop	de				; alternate
	exx					; 
	pop	hl				; 
	pop	de				; 
	jr	stringFindLoop			; 