; This program is free software. It comes without any warranty, to
; the extent permitted by applicable law. You can redistribute it
; and/or modify it under the terms of the Do What The Fuck You Want
; To Public License, Version 2, as published by Sam Hocevar. See
; http://sam.zoy.org/wtfpl/COPYING for more details.

documentSection("KEYBOARD")
document("It might be nice to have a GetString routine at some point.  Whatever.")

;------ GetAscii ---------------------------------------------------------------
GetAscii:
documentRoutine("GetAscii")
document("Gets a single key press, with modifiers from shiftFlags processed.  2nd and")
document("Alpha flags are reset, unless alpha-lock is enabled, in which case only 2nd is")
document("reset.")
document("Inputs:")
document(" - inputerFlags")
document("Outputs:")
document(" - ASCII char in B, might be a control char")
document(" - D: Shift flags")
document(" - E: inputerFlags")
document("Destroys:")
document(" - Pretty much everything, most likely.")
	ld	a, (mflags + inputerFlags)
	and	allowAllMask
	ret	z
	call	GetKeyShifts
	; Test for A-Z
	bit	allowAlpha, e
	jr	z, getAsciiNum
	bit	shiftAlpha, d
	jr	z, getAsciiNum
	; Alpha A-Z
	ld	hl, textAlphaTable
	call	getAsciiScanTable
	jr	nz, +_
	bit	allowLwrAlpha, e
	jr	z, ClearShiftFlags		; implicit return
	bit	shiftLwrAlpha, d
	jr	z, ClearShiftFlags		; implicit return
	; Alpha a-z
	ld	a, 20h
	add	a, b
	ld	b, a
	jr	ClearShiftFlags			; implicit return
_:	ld	hl, textAlphaSymbolTable
	bit	allowSymbols, e
	jr	z, GetAscii
	call	getAsciiScanTable
	jr	z, ClearShiftFlags		; implicit return
	jr	getAsciiControlChar
getAsciiNum:
	; Numbers 0-9
	ld	hl, textNumberSymbolTable
	bit	allowNumbers, e
	jr	z, getAsciiControlChar
	ld	hl, textNumberTable
	call	getAsciiScanTable
	jr	z, ClearShiftFlags		; implicit return
	ld	hl, textNumberSymbolTable
	bit	allowSymbols, e
	call	nz, getAsciiScanTable
	jr	z, ClearShiftFlags		; implicit return
getAsciiControlChar:
	bit	allowControl, e
	jr	z, GetAscii
	bit	shift2nd, d
	jr	z, +_
	ld	hl, text2ndControlCharTable
getAsciiCtrl:
	call	getAsciiScanTable
	jr	z, ClearShiftFlags		; implicit return
	jr	GetAscii
_:	bit	shiftAlpha, d
	jr	z, +_
	ld	hl, textAlphaControlCharTable
	jr	getAsciiCtrl
_:	ld	hl, textControlCharTable
	jr	getAsciiCtrl

	
getAsciiScanTable:
; Scans a table mapping an input in A to an output in B.
; Input:
;  - HL: Pointer to struct specifying entries in real table and table address.
; Outputs:
;  - B: Code
;  - NZ on failure
; Destroys:
;  - C
;  - HL
	ld	b, (hl)
	inc	hl
	ld	c, (hl)
	inc	hl
	ld	h, (hl)
	ld	l, c
_:	cp	(hl)
	jr	z, +_
	inc	hl
	inc	hl
	djnz	-_
	cp	b
	ret
_:	inc	hl
	ld	b, (hl)
	cp	a
	ret


;------ ClearShiftFlags --------------------------------------------------------
ClearShiftFlags:
documentRoutine("ClearShiftFlags")
document("Clears 2nd and Alpha.  However, if Alpha Lock is on, alpha is not cleared.  If")
document("a special cursor is selected, it is not cleared.  See ClearAllShiftFlags for a")
document("routine to clear all shift flags.")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Flags cleared")
document("Destroys:")
document(" - AF")
	ld	a, (mflags + shiftFlags)
	bit	shiftOther, a
	ret	nz
	and	shift2ndMask ^ 255
	ld	(mflags + shiftFlags), a
	bit	shiftALock, a
	ret	nz
	and	(shiftLwrAlphaMask | shiftAlphaMask) ^ 255
	ld	(mflags + shiftFlags), a
	ret


;------ ClearAllShiftFlags -----------------------------------------------------
ClearAllShiftFlags:
documentRoutine("ClearAllShiftFlags")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Flags cleared")
document("Destroys:")
document(" - AF")
	ld	a, (mflags + shiftFlags)
	and	(shiftALockMask | shiftLwrAlphaMask | shiftAlphaMask | shift2ndMask | shiftInsertMask | shiftOtherMask) ^ 255
	ld	(mflags + shiftFlags), a
	ret


;------ GetKeyShifts -----------------------------------------------------------
GetKeyShifts:
documentRoutine("GetKeyShifts")
document("Gets a single key press.  If the cursor is enabled, it will blink, and change")
document("according to shiftFlags.")
document("Input:")
document(" - shiftFlags you want preset")
document("Output:")
document(" - Key code in A")
document(" - Flags")
document("Destroys:")
document(" - AF")
document(" - BC")
document(" - DE")
document(" - HL")
getAsciiLoop:
;	di
;	ld	bc, (cursorTimerPeriod)
;	ld	(cursorTimer), bc
;	ei
	call	GetKey
	ld	l, a
	ld	a, (mflags + inputerFlags)	; E: inputerFlags
	ld	e, a
	ld	a, (mflags + shiftFlags)	; D: shiftFlags
	ld	d, a
	ld	a, l
	; Check cursor options
	cp	skAlpha
	jr	nz, getKeyNoAlpha
	bit	allowAlpha, e
	jr	z, getAsciiLoop
	ld	a, d
	bit	shift2nd, a
	jr	z, +_
	and	shift2ndMask ^ 255
	or	shiftAlphaMask | shiftALockMask
	jr	getAsciiSetFlags
_:	bit	shiftAlpha, a
	jr	nz, +_
	or	shiftAlphaMask
	and	shiftLwrAlphaMask ^ 255
	jr	getAsciiSetFlags
_:	bit	allowLwrAlpha, e
	jr	nz, ++_
_:	and	(shiftAlphaMask | shiftLwrAlphaMask) ^ 255
	jr	getAsciiSetFlags
_:	bit	shiftLwrAlpha, a
	jr	nz, --_
	or	shiftLwrAlphaMask	
getAsciiSetFlags:
	ld	(mflags + shiftFlags), a
	jr	getAsciiLoop
getKeyNoAlpha:
	cp	sk2nd
	ret	nz
	ld	a, d
	bit	shift2nd, a
	jr	nz, +_
	or	shift2ndMask
	jr	getAsciiSetFlags
_:	and	shift2ndMask ^ 255
	jr	getAsciiSetFlags


;------ Hex input --------------------------------------------------------------
GetHexByte:
documentRoutine("GetHexByte")
document("Do not attempt to get a byte across a line break.")
document("Gets byte at current location.  Returns byte in A")
document("Registers:")
document("b: returns nibble from getNibble")
document("d: byte")
	push	bc
	push	de
	push	hl
	call	CursorOn
	call	getNibble
	sla	b
	sla	b
	sla	b
	sla	b
	push	bc
	call	getNibble
	ld	a, b
	pop	bc
	add	a, b
	push	af
	call	CursorOff
	pop	af
	pop	hl
	pop	de
	pop	bc
	ret
getNibble:
	call	GetKey
	ld	hl, hexKeyTable
	ld	b, hexKeyTableSize
_:	cp	(hl)
	jr	z, foundNibble
	inc	hl
	djnz	-_
	jr	getNibble
foundNibble:
	dec	b
	ld	a, '0'
	add	a, b
	cp	3Ah
	jr	c, +_
	add	a, 7
_:	call	PutC
	ret


;------ GetKey -----------------------------------------------------------------
GetKey:
documentRoutine("GetKey")
document("Waits for a keypress, and returns it.")
document("This may APD.")
document("Interrupts must be enabled.")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Scan code in A")
document("Destroys:")
document(" - Nothing")
	push	bc
	push	af
	ld	bc, (cursorTimerPeriod)
	ld	(cursorTimer), bc
	ld	a, (mflags + curFlags)
	bit	curAble, a
	jr	z, +_
	or	curFlashMask
	ld	(mflags + curFlags), a
_:	pop	af
	pop	bc
getKeyLoop:
	call	GetCSC
	or	a
	jr	nz, EraseCursor
	ei
	halt
	ld	a, (mflags + curFlags)
	bit	curAble, a
	jr	z, getKeyLoop
	bit	curFlash, a
	jr	z, getKeyLoop
	call	CursorToggle
	jr	getKeyLoop
EraseCursor:
	push	af
	ld	a, (mflags + curFlags)
	bit	curAble, a
	jr	z, +_
	bit	curOn, a
	call	nz, CursorToggle
_:	pop	af
	ret
CursorToggle:
	push	bc
	push	de
	push	hl
	ld	a, (mflags + curFlags)
	and	curFlashMask ^ 255
	bit	curOn, a
	jr	nz, cursorErase
cursorShow:
	or	curOnMask
	ld	(mflags + curFlags), a
	call	ReadChar
	ld	(cursorChar), a
	call	FixCursor
	ld	a, (mflags + shiftFlags)
	ld	c, 0
	ld	b, a
	ld	a, chCur
	bit	shiftOther, b
	jr	z, cursorNormalSet
	; Handle other cases
	ld	a, chCurFull
	bit	shiftFull, b
	jr	nz, cursorShowCursor
;	ld	a, 
;	bit	shiftBox, b
	inc	a
	bit	shiftVertLine, b
	jr	nz, cursorShowCursor
	inc	a
	bit	shiftUnderline, b
	jr	nz, cursorShowCursor
	ld	a, chBox
	bit	shiftBox, b
	jr	nz, cursorShowCursor
	ld	a, '?'
	jr	cursorShowCursor
cursorNormalSet:
	bit	shiftInsert, b
	jr	z, +_
	add	a, 4
_:	bit	shift2nd, b
	jr	z, +_
	inc	a
	jr	cursorShowCursor
_:	bit	shiftAlpha, b
	jr	z, cursorShowCursor
	add	a, 2
	bit	shiftLwrAlpha, b
	jr	z, cursorShowCursor
	inc	a	
cursorShowCursor:
	ld	hl, (currentRow)
	push	hl
	call	PutC
	;call	BackspaceCursor
	pop	hl
	call	Locate
	jr	cursorToggleRet
cursorErase:
	and	curOnMask ^ 255
	ld	(mflags + curFlags), a
	; Erase the cursor
#if textInverse != 0
	.error	"Change cursorErase to know the correct value of textInverse"
#endif
	ld	a, (cursorChar)
	and	80h
	rlc	a
	ld	b, a
	ld	a, (mflags + textFlags)
	ld	c, a
	and	textInverseMask^255
	or	b
	ld	(mflags + textFlags), a
	ld	a, (cursorChar)
	and	7Fh
	ld	hl, (currentRow)
	push	hl
	call	PutC
	pop	hl
	call	Locate
	;call	BackspaceCursor
	ld	a, c
	ld	(mflags + textFlags), a
cursorToggleRet:
	pop	hl
	pop	de
	pop	bc
	ret


;------ CursorOn ---------------------------------------------------------------
CursorOn:
documentRoutine("CursorOn")
document("Enables the blinking cursor that GetKey can show")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Documented effect")(s)
document("Destroys:")
document(" - Nothing")
	push	af
	ld	a, (mflags + curFlags)
	or	curAbleMask | curFlashMask
;	and	curOnMask ^ 255
	ld	(mflags + curFlags), a
;	call	CursorToggle;cursorShow
	pop	af
	ret


;------ CursorOff --------------------------------------------------------------
CursorOff:
documentRoutine("CursorOff")
document("Disables the blinking cursor")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Documented effect")(s)
document("Destroys:")
document(" - Nothing")
	push	af
	ld	a, (mflags + curFlags)
	and	(curAbleMask) ^ 255
	ld	(mflags + curFlags), a
;	call	EraseCursor
	pop	af
	ret


;------ GetCSC -----------------------------------------------------------------
GetCSC:
documentRoutine("GetCSC")
document("Returns the most recent keypress and removes it from the key queue.")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Scan code in A")
document("Destorys:")
document(" - Nothing")
	ld	a, (keyBuffer)
	push	af
	xor	a
	ld	(keyBuffer), a
	pop	af
	ret


;------ PeekCSC ----------------------------------------------------------------
PeekCSC:
documentRoutine("PeekCSC")
document("Returns the most recent keypress but leaves it in the key queue.")
document("This is a pretty dumb routine.")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Scan code in A")
document("Destroys:")
document(" - Nothing")
	ld	a, (keyBuffer)
	ret


;------ Scan Keyboard ----------------------------------------------------------
;	.dw	scanKeyboardEnd-ScanKeyboardSource
ScanKeyboard:
documentRoutine("ScanKeyboard")
document("Implements the core logic used for keyboard scanning, including timing for")
document("debouncing.")
document("You can call this manually if interrupts are disabled, but you will need to do")
document("so at about 50 Hz or the debouncing timing will be strange.")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Documented effect (s)")
document("Destroys:")
document(" - AF")
document(" - BC")
	call	RawGetCSC
; Key repeat logic:
; Key must be held for at least minKeyHoldTime before press registers.
; Thereafter, the same key code will not be accepted after release for keyBlockTime.
; As long as the key is being held, reissue the keypress to keyBuffer every keyRepeatWaitTime.
	; Any code?
	or	a
	jr	nz, dontReleaseKey
	ld	a, (lastKey)
	or	a
	ret	z	;jr	z, noKeys
	; Otherwise, begin blocking timer
	ld	a, (lastKeyBlockTimer)
	dec	a
	ld	(lastKeyBlockTimer), a
	ret	nz	;jr	nz, nokeys
	; The timer has expired, so allow the key again
	ld	(lastKey), a	; already 0
	ret
dontReleaseKey:
	ld	c, a	; Save code for a moment
	; Reset the blocking timer
	ld	a, keyBlockTime
	ld	(lastKeyBlockTimer), a
	; Is the registered key the same as the last key?
	; This also covers the 0 case.
	ld	a, (lastKey)
	cp	c
	jr	nz, newKey	; If not, accept the new key
	; It's the same old key, so decrement the hold timer
	ld	a, (lastKeyHoldTimer)
	or	a
	jr	nz, decHoldTimer
	; Wait, the key has already been accepted.
	; Check if we're already in repeat mode, or need to wait
	ld	a, (lastKeyFirstRepeatTimer)
	or	a
	jr	z, +_	; Repeat mode
	dec	a
	ld	(lastKeyFirstRepeatTimer), a
	ret	;jr	noKeys
_:	; OK, so we're in repeat mode.  Now do the reapeat timer.
	ld	a, (lastKeyRepeatTimer)
	or	a
	jr	z, acceptKeypress	; The key was previously accepted, AND it's now ready to be reissued.
	dec	a
	ld	(lastKeyRepeatTimer), a
	ret
decHoldTimer:
	dec	a
	ld	(lastKeyHoldTimer), a
	jr	z, acceptKeypress
	; This key isn't ripe yet, so ignore it for now.
	ret	;jr	noKeys
newKey:	; Save the key code
	ld	a, c
	ld	(lastKey), a
	; Reset timers
	ld	bc, keyBlockTime*256 + minKeyHoldTime
	ld	(lastKeyHoldTimer), bc
	ld	bc, keyRepeatWaitTime*256 + keyFirstRepeatWaitTime
	ld	(lastKeyFirstRepeatTimer), bc
	; Now RET, and leave decrementing the debouncing timers for the next
	; interrupt cycle
	ret	;jr	noKeys
acceptKeypress:
	; Reset timers
	ld	a, keyBlockTime
	ld	(lastKeyBlockTimer), a
	ld	a, keyRepeatWaitTime
	ld	(lastKeyRepeatTimer), a
	ld	bc, suspendDelay
	ld	(suspendTimer), bc
	; Now write the key to the buffer
	ld	a, (lastKey)
	ld	(keyBuffer), a
;noKeys:	ret
	ret
scanKeyboardEnd:


;------ GetCSC -----------------------------------------------------------------
;	.dw	rawGetCSCEnd-RawGetCSCSource
;RawGetCSCSource:
RawGetCSC:
documentRoutine("RawGetCSC")
document("Scans the keyboard matrix for any pressed key, returning the first it finds,")
document("or 0 if none.")
document("Inputs:")
document(" - None")
document("Output:")
document(" - Code in A, or 0 if none")
document("Destroys:")
document(" - BC")
	ld	a, 0FFh
	out	(pKey), a
	pop	af
	push	af
	ld	bc, 07BFh
_:	; Matrix scan loop
	ld	a, c
	out	(pKey), a
	rrca
	ld	c, a
	pop	af	; Probably should waste at least 20 cycles here.
	push	af	
;	push	ix
;	pop	ix
	in	a, (pKey)
	cp	0ffh
	jr	nz, +_	; Any key pressed?
	djnz	-_
	; No keys pressed in any key group, so return 0.
	xor	a
	ret
_:	; Yay! Found a key, now form a scan code
	dec	b
	sla	b
	sla	b
	sla	b
	; Get which bit in A is reset
_:	rrca
	inc	b
	jr	c, -_
	ld	a, b
	ret
rawGetCSCEnd: